/*
 * $Id: Cc.java,v 1.2 2003/05/28 15:22:46 kenta Exp $
 *
 * Copyright (c) 2003 Kenta Cho. All rights reserved.
 */
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Calendar;
import java.util.Random;

import javax.microedition.io.Connector;

import com.nttdocomo.device.Camera;
import com.nttdocomo.lang.IllegalStateException;
import com.nttdocomo.opt.device.Camera2;
import com.nttdocomo.opt.ui.j3d2.C3DGraphics;
import com.nttdocomo.system.InterruptedOperationException;
import com.nttdocomo.ui.Canvas;
import com.nttdocomo.ui.Display;
import com.nttdocomo.ui.Frame;
import com.nttdocomo.ui.Graphics;
import com.nttdocomo.ui.Image;
import com.nttdocomo.ui.MApplication;
import com.nttdocomo.ui.MediaImage;
//import com.nttdocomo.ui.MediaManager;
import com.nttdocomo.ui.UIException;

/**
 * CameraClock main class.
 */
public class Cc extends MApplication {
	private SimpleC3DCanvas canvas;
	private C3DGraphics g;
	private int scWidth, scHeight;

	private final int fx1 = 65536;

	private boolean conc;
	// 0 - Show clock.  1 - Capture a picture.
	private int state;

	/**
	 * Initialize.
	 */
	private void init() {
		canvas = new SimpleC3DCanvas();
		scWidth = canvas.getWidth();
		scHeight = canvas.getHeight();
		changeState(0);
		pccv = new SimpleCanvas();
	}

	private void setSoftLabels() {
		canvas.setSoftLabel(Frame.SOFT_KEY_1, "Be");
		if (conc) {
			canvas.setSoftLabel(Frame.SOFT_KEY_2, "Ҏ");
		} else {
			canvas.setSoftLabel(Frame.SOFT_KEY_2, "I");
		}
		pccv.setSoftLabel(Frame.SOFT_KEY_1, "捞");
		pccv.setSoftLabel(Frame.SOFT_KEY_2, "~");
	}

	private void clearSoftLabels() {
		pccv.setSoftLabel(Frame.SOFT_KEY_1, null);
		pccv.setSoftLabel(Frame.SOFT_KEY_2, null);
	}

	/**
	 * Init a canvas after the C3DGraphics instance becomes available.
	 */
	private void initCanvas()
		throws UIException, InterruptedOperationException, IOException {
		int lt = getLaunchType();
		if ( lt == LAUNCHED_AS_CONCIERGE ) {
			conc = true;
		} else {
			conc = false;
		}
		setSoftLabels();

		g.set3DMode(true);
		g.viewport(0, 0, scWidth, scHeight);
		g.clearColor(0, 0, 0, fx1);

		g.enable(C3DGraphics.C3DL_TEXTURE_2D);
		initTexture(10);
		try {
			loadTexture(10);
		} catch (IOException e) {
		}

		g.blendFunc(
			C3DGraphics.C3DL_SRC_ALPHA,
			C3DGraphics.C3DL_ONE_MINUS_SRC_ALPHA);
		g.enable(C3DGraphics.C3DL_BLEND);
		g.enable(C3DGraphics.C3DL_LINE_SMOOTH);
		g.disable(C3DGraphics.C3DL_DEPTH_TEST);
		g.disable(C3DGraphics.C3DL_CULL_FACE);

		g.matrixMode(C3DGraphics.C3DL_PROJECTION);
		g.loadIdentity();
		int aspect = fx1 * scWidth / scHeight;
		g.perspective(45 * fx1, aspect, fx1 / 256, 512 * fx1);

		g.matrixMode(C3DGraphics.C3DL_MODELVIEW);
		g.loadIdentity();

		g.disable(C3DGraphics.C3DL_LIGHTING);

		initViewpoint();
	}

	private void changeState(int st) {
		state = st;
		switch (state) {
			case 0 :
			Display.setCurrent(canvas);
				break;
			case 1 :
				Display.setCurrent(pccv);
				break;
		}
	}

	/*
	 * Capture a picture and convert to the texture.
	 */
	private MediaImage mi;
	private Image pi;

	private final int pw = 128, ph = 128;

	private Canvas pccv;
	private Graphics pcg;
	private int pst;
	private int pstn[] =
		{
			Graphics.FLIP_NONE,
			Graphics.FLIP_ROTATE_LEFT,
			Graphics.FLIP_ROTATE,
			Graphics.FLIP_ROTATE_RIGHT };

	/**
	 * Handle a camera.
	 */
	private void takePicture()
		throws InterruptedOperationException, UIException, IOException {
		Camera camera = Camera.getCamera(0);
		camera.setImageSize(288, 352);
		camera.setFrameImage(null);
		camera.setAttribute(Camera.DEV_QUALITY, Camera.ATTR_QUALITY_HIGH);
		camera.setAttribute(Camera2.DEV_IMAGE_ENCODER, Camera2.ATTR_RAW_IMAGE_ENCODER);
		camera.takePicture();
		mi = camera.getImage(0);
		//mi = MediaManager.getImage("resource:///test.gif");
		mi.use();
		pi = mi.getImage();
		pst = 0;
	}

	private void showPictureCheck() {
		changeState(1);
		pcg = pccv.getGraphics();
		pcg.lock();
		pcg.setFlipMode(pstn[pst]);
		pcg.drawScaledImage(
			pi,
			0,
			0,
			pw,
			ph,
			0,
			(pi.getHeight() - pi.getWidth()) / 2,
			pi.getWidth(),
			pi.getWidth());
		pcg.clearRect(24, 160, 200, 100);
		pcg.drawString("̎ʐ^荞݂܂H", 24, 200);
		pcg.drawString("EL[Ŏʐ^]ł܂", 24, 224);
		pcg.unlock(true);
	}

	private void setTexture(int n, int w, int h, byte[] bp) {
		g.bindTexture(C3DGraphics.C3DL_TEXTURE_2D, n);
		g.texImage2D(
			C3DGraphics.C3DL_TEXTURE_2D,
			0,
			C3DGraphics.C3DL_RGBA8,
			w,
			h,
			0,
			C3DGraphics.C3DL_RGBA,
			C3DGraphics.C3DL_UNSIGNED_BYTE,
			bp);
	}

	//private boolean reloadTex = false;
	
	private void makeTextureFromPicture(int n) throws IOException {
		pcg.clearRect(24, 160, 200, 100);
		pcg.drawString("ʐ^荞ł܂", 24, 200);
		pcg.drawRect(24, 210, 180, 20);
		clearSoftLabels();
		byte[] bp = new byte[pw * ph * 4];
		int i = 0;
		for (int y = 0; y < ph; y++) {
			for (int x = 0; x < pw; x++, i += 4) {
				int p = pcg.getRGBPixel(x, y);
				bp[i] = (byte) ((p & 0xff0000) >> 16);
				bp[i + 1] = (byte) ((p & 0xff00) >> 8);
				bp[i + 2] = (byte) (p & 0xff);
				bp[i + 3] = (byte) 255;
			}
			if ((y & 15) == 0) {
				pcg.fillRect(24, 210, 170 * y / ph, 20);
			}
		}
		OutputStream out = Connector.openOutputStream("scratchpad:///0");
		out.write(bp);
		out.close();
		mi.dispose();
		setSoftLabels();
		changeState(0);
		setTexture(n, pw, ph, bp);
	}

	private void loadTexture(int n) throws IOException {
		byte[] bp = new byte[pw * ph * 4];
		InputStream in = Connector.openInputStream("scratchpad:///0");
		int l = in.read(bp);
		in.close();
		if (l == -1)
			return;
		setTexture(n, pw, ph, bp);
	}

	private void initTexture(int n) {
		byte[] bp = new byte[pw * ph * 4];
		setTexture(n, pw, ph, bp);
	}

	// Viewpoint data.
	private int ex, ey, ez, eax, eay, eaz;
	private int sx, sy, sz, sax, say, saz;

	/**
	 * Move an eye position and a looking position.
	 */
	private void moveViewpoint() {
		ex += (eax - ex) >> 5;
		ey += (eay - ey) >> 5;
		ez += (eaz - ez) >> 5;
		sx += (sax - sx) >> 7;
		sy += (say - sy) >> 7;
		sz += (saz - sz) >> 7;
	}

	private void initViewpoint() {
		sy = say = fx1 / 6;
		ez = eaz = fx1 * 3;
	}

	private void setViewpointEye() {
		eax = randNS(fx1 / 2 * 3);
		eay = randNS(fx1 / 2 * 3);
		eaz = fx1 / 2 * 5 + randN(fx1 / 2 * 3);
	}

	private void setViewpointSee() {
		sax = randNS(fx1 / 3);
		say = randNS(fx1 / 3) + fx1 / 6;
		saz = randNS(fx1 / 3);
	}

	/**
	 * set current time(hour/minute/second)
	 */
	private void setTime() {
		Calendar calendar = Calendar.getInstance();
		int hr = calendar.get(Calendar.HOUR),
			mn = calendar.get(Calendar.MINUTE),
			sc = calendar.get(Calendar.SECOND);
		hd[0] = hr * (360 * fx1 / 12) + mn * (360 * fx1 / 12 / 60);
		hd[1] = mn * (360 * fx1 / 60) + sc * (360 * fx1 / 60 / 60);
		hd[2] = sc * (360 * fx1 / 60);
	}

	private int cnt = 0;

	/**
	* Move 1 frame.
  */
	private void move() {
		if (state != 0)
			return;
		moveViewpoint();
		setTime();
		switch (cnt & 63) {
			case 0 :
				setViewpointEye();
				break;
			case 32 :
				setViewpointSee();
				break;
		}
		cnt++;
	}

	private boolean prsd = true;

	/**
	 * Check a key pad state.
	 */
	public void handleKeyState(int key)
		throws UIException, IOException, InterruptedOperationException {
		if ( key == 0 ) {
			prsd = false;
		} else {
			if ( prsd ) return;
			prsd = true;
		}
		switch (key) {
			case (1<<Display.KEY_SOFT1) :
				switch (state) {
					case 0 :
						takePicture();
						showPictureCheck();
						break;
					case 1 :
						makeTextureFromPicture(10);
						break;
				}
				break;
			case (1<<Display.KEY_SOFT2) :
				switch (state) {
					case 0 :
						if (conc) {
							deactivate();
						} else {
							terminate();
						}
						break;
					case 1 :
						changeState(0);
						break;
				}
				break;
			case (1<<Display.KEY_LEFT) :
				if (state == 1) {
					pst--;
					pst &= 3;
					showPictureCheck();
					break;
				}
				break;
			case (1<<Display.KEY_RIGHT) :
				if (state == 1) {
					pst++;
					pst &= 3;
					showPictureCheck();
					break;
				}
				break;
		}
	}

	/*
	 * Draw a 3D clock.
	 */
	private int[] hd = new int[3];
	private int[] hl = { fx1 / 8 * 5, fx1 / 8 * 7, fx1 / 8 * 6 };
	private int[][] hc = { { fx1, fx1 / 3 * 2, fx1 / 3 * 2 }, {
			fx1, fx1, fx1 / 3 * 2 }, {
			fx1 / 3 * 2, fx1, fx1 / 3 * 2 }
	};

	private void drawPlate(int z, int a) {
		g.bindTexture(C3DGraphics.C3DL_TEXTURE_2D, 10);
		g.color(fx1, fx1, fx1, a);
		g.begin(C3DGraphics.C3DL_TRIANGLE_FAN);
		g.texCoord(0, 0);
		g.vertex(-fx1, fx1, z);
		g.texCoord(fx1, 0);
		g.vertex(fx1, fx1, z);
		g.texCoord(fx1, fx1);
		g.vertex(fx1, -fx1, z);
		g.texCoord(0, fx1);
		g.vertex(-fx1, -fx1, z);
		g.end();
	}

	private int[][] spp = { { -1, 1 }, {
			1, 1 }, {
			1, -1 }, {
			-1, -1 }
	};

	private void drawPlateSide(int z, int a) {
		int i, ni;
		g.color(fx1, fx1, fx1, a);
		g.begin(C3DGraphics.C3DL_QUADS);
		for (i = 0, ni = 1; i < 4; i++, ni++) {
			if (ni >= 4)
				ni = 0;
			g.vertex(fx1 * spp[i][0], fx1 * spp[i][1], 0);
			g.vertex(fx1 * spp[ni][0], fx1 * spp[ni][1], 0);
			g.vertex(fx1 * spp[ni][0], fx1 * spp[ni][1], z);
			g.vertex(fx1 * spp[i][0], fx1 * spp[i][1], z);
		}
		g.end();
		g.color(fx1, fx1, fx1, fx1 / 3 * 2);
		g.begin(C3DGraphics.C3DL_LINE_LOOP);
		for (i = 0; i < 4; i++) {
			g.vertex(fx1 * spp[i][0], fx1 * spp[i][1], 0);
		}
	}

	private void drawPoint(int cx, int cy, int z, int a) {
		int i, ni;
		int w = fx1 / 16;
		g.color(fx1, fx1, fx1, a);
		g.begin(C3DGraphics.C3DL_QUADS);
		for (i = 0, ni = 1; i < 4; i++, ni++) {
			if (ni >= 4)
				ni = 0;
			g.vertex(cx + w * spp[i][0], cy + w * spp[i][1], 0);
			g.vertex(cx + w * spp[ni][0], cy + w * spp[ni][1], 0);
			g.vertex(cx + w * spp[ni][0], cy + w * spp[ni][1], z);
			g.vertex(cx + w * spp[i][0], cy + w * spp[i][1], z);
		}
		g.vertex(cx + w * spp[0][0], cy + w * spp[0][1], 0);
		g.vertex(cx + w * spp[1][0], cy + w * spp[1][1], 0);
		g.vertex(cx + w * spp[2][0], cy + w * spp[2][1], 0);
		g.vertex(cx + w * spp[3][0], cy + w * spp[3][1], 0);
		g.end();
	}

	private void drawHand(int i, int zf, int a) {
		g.pushMatrix();
		g.rotate(hd[i], 0, 0, -fx1);
		int l = hl[i],
			w = fx1 / 40 * (3 - i),
			z = i * fx1 / 6 * zf,
			sl = fx1 / 8;
		g.begin(C3DGraphics.C3DL_TRIANGLE_FAN);
		g.color(hc[i][0], hc[i][1], hc[i][2], a);
		g.vertex(-w, l, z);
		g.vertex(w, l, z);
		g.color(hc[i][0] / 2, hc[i][1] / 2, hc[i][2] / 2, a);
		g.vertex(w, sl, z);
		g.vertex(-w, sl, z);
		g.end();
		if (zf == 1) {
			g.begin(C3DGraphics.C3DL_LINE_LOOP);
			g.color(hc[i][0], hc[i][1], hc[i][2], fx1);
			g.vertex(-w, l, z);
			g.vertex(w, l, z);
			g.vertex(w, sl, z);
			g.vertex(-w, sl, z);
			g.end();
		}
		g.popMatrix();
	}

	private void drawClock() {
		int cw = fx1 / 5;
		g.enable(C3DGraphics.C3DL_TEXTURE_2D);
		drawPlate(-cw, fx1 / 4 * 3);

		g.disable(C3DGraphics.C3DL_TEXTURE_2D);
		drawPoint(0, -fx1 / 6 * 5, -cw, fx1 / 2);
		drawPoint(0, fx1 / 6 * 5, -cw, fx1 / 2);
		drawPoint(-fx1 / 6 * 5, 0, -cw, fx1 / 2);
		drawPoint(fx1 / 6 * 5, 0, -cw, fx1 / 2);
		for (int i = 0; i < 3; i++)
			drawHand(i, -1, fx1 / 2);
		drawPlateSide(-fx1 / 2, fx1 / 5);

		g.enable(C3DGraphics.C3DL_TEXTURE_2D);
		drawPlate(0, fx1 / 8);

		g.disable(C3DGraphics.C3DL_TEXTURE_2D);
		for (int i = 0; i < 3; i++)
			drawHand(i, 1, fx1 / 16 * 15);
	}

	private void drawBack() {
		g.enable(C3DGraphics.C3DL_TEXTURE_2D);
		g.bindTexture(C3DGraphics.C3DL_TEXTURE_2D, 10);
		g.color(fx1 / 4, fx1 / 4, fx1 / 2, fx1);
		g.begin(C3DGraphics.C3DL_TRIANGLE_FAN);
		int z = -fx1 * 13, s = 20;
		g.texCoord(0, 0);
		g.vertex(-fx1 * s, fx1 * s, z);
		g.texCoord(fx1, 0);
		g.vertex(fx1 * s, fx1 * s, z);
		g.texCoord(fx1, fx1);
		g.vertex(fx1 * s, -fx1 * s, z);
		g.texCoord(0, fx1);
		g.vertex(-fx1 * s, -fx1 * s, z);
		g.end();
	}

	/**
	 * Draw to the canvas.
	 */
	private void draw() {
		if (state != 0)
			return;
		g.lock();
		g.clear(C3DGraphics.C3DL_COLOR_BUFFER);
		g.pushMatrix();
		g.lookAt(ex, ey, ez, sx, sy, sz, 0, fx1, 0);
		drawBack();
		drawClock();
		g.popMatrix();
		g.flush();
		g.unlock(true);
	}

	/**
	 * Main loop. (10 frames/sec)
	 */
	public void run() {
		long prvTicks, nowTicks;
		int frame;
		prvTicks = System.currentTimeMillis();
		for (;;) {
			nowTicks = System.currentTimeMillis();
			frame = (int) (nowTicks - prvTicks) / 100;
			if (frame <= 0) {
				frame = 1;
				prvTicks += 100;
				//while ( System.currentTimeMillis() < prvTicks ) {}
				try {
					Thread.sleep(prvTicks - System.currentTimeMillis());
				} catch (InterruptedException e) {
				}
			} else {
				prvTicks = nowTicks;
				if (frame > 4) {
					frame = 4;
				}
			}
			if (g == null) {
				g = canvas.getC3DGraphics();
				if (g != null) {
					try {
						initCanvas();
					} catch (Exception e) {
						continue;
					}
					setTime();
				} else {
					continue;
				}
			}
			try {
				handleKeyState(canvas.getKeypadState());
			} catch (Exception e) {
			}
			for (; frame > 0; frame--) {
				move();
			}
			draw();
		}
	}

	/*
	 * Random util methods.
	 */
	private static Random rnd = new Random();

	private int randN(int n) {
		return Math.abs(rnd.nextInt() % n);
	}

	private int randNS(int n) {
		return rnd.nextInt() % n;
	}

	/**
	 * Sleep when the phone is folded.
	 */
	public void processSystemEvent(int type, int param) {
		if (type == FOLD_CHANGED_EVENT && param == 0) {
			try {
				sleep();
			} catch ( IllegalStateException e ) {
				deactivate();
				sleep();
			}
		}
	}

	/**
	 * Bootstrap.
	 */
	public void start() {
		init();
		run();
	}
}
